apc.indiv.logit.TS <- function(data, dep.var, covariates=NULL, maxit.loop=10, 
                               maxit.linesearch=30, tolerance=.002, init="ols",
                               inv.tol=NULL, d1.tol=.002, 
                               custom.kappa=NULL, custom.zeta=NULL){
  ##################
  # This function breaks down the logit estimator step-by-step to exploit the 
  # identity structure of the TS model. Y is the dependent variable, mu the 
  # linear predictor of form mu = kappa*T + zeta*Z where T are the time 
  # dummies and Z are the covariates. 
  ##################  
  if(is.null(inv.tol)){
    inv.tol <- .Machine$double.eps
  }  
  
  if(!is.null(covariates)){
    
    #### Calculate things that will be used repeatedly ####
    
    ## get Y and Z
    Y <- as.matrix(data[dep.var])
    Z <- as.matrix(data[names(data) %in% covariates])
    
    ## get cell name
    # add cell.name to data: age-cohort cell identifier (if not already there)
    if (!"cell.name" %in% colnames(data)) {
      cell.name <- paste("ik", as.character(data$age), as.character(data$cohort),
                         sep="_")
      data <- cbind(cell.name, data)
    } else {
      warning("Variable cell.name must be unique identifier of age-cohort cells;
              if preexisting variable cell.name has alternate meaning please 
              rename")
    }
    
    ## get first part of first derivative
    cell.sum.Y.withname <- ddply(data, .variables=c("cell.name"),
                                 function(subdata, colnames){colSums(subdata[colnames])},
                                 dep.var)
    colnames(cell.sum.Y.withname)[2:ncol(cell.sum.Y.withname)] <- 
      paste("cell.sum", dep.var, sep=".")
    cell.sum.Y <- cell.sum.Y.withname[2:ncol(cell.sum.Y.withname)]
    
    d1e1_top <- as.matrix(cell.sum.Y)
    d1e1_bot <- t(Z) %*% Y
    
    #### Get initial starting values ####
    
    if((!is.null(custom.kappa)|!is.null(custom.zeta)) & !(init=="custom"))
      warning("custom.kappa and custom.zeta are not used without init=custom")
    
    if (init == "ols"){
      # use estimate.TS once to get estimated kappa and zeta
      initial.TS <- apc.indiv.estimate.TS(data, dep.var = dep.var, 
                                          covariates = covariates)
      kappa <- as.matrix(initial.TS$kappahat)
      zeta <- as.matrix(initial.TS$zetahat) 
    } else if (init == "zero"){
      zeta <- rep(0, ncol(Z)) 
      kappa <- rep(0, nrow(cell.sum.Y))
    } else if (init == "custom"){
      zeta <- custom.zeta
      kappa <- custom.kappa
    } else stop("init must be one of 'ols', 'zero', 'custom'")
    
    # set loop iteration counter
    i <- 1
    
    # get mu, the individual predictor
    kappa.withname <- cbind(cell.sum.Y.withname[1], kappa)
    indiv.cell.name <- data["cell.name"]
    indiv.kappa <- join(indiv.cell.name, kappa.withname, by="cell.name") 
    
    mu <- indiv.kappa$kappa + Z %*% zeta
    colnames(mu) <- "mu"
    
    # evaluate the log-likelihood at the estimated values of kappa and zeta: 
    lik <- (t(Y) %*% mu) - sum(log(1+exp(mu)))
    
    # --------------------------
    
    #### The loop ####
    
    while (i <= maxit.loop){
      
      ## Step 1: Elements of updating ##
      
      # get Nx1 vector of probabilities
      pi <- exp(mu)/(1+exp(mu))
      colnames(pi) <- "pi"
      
      # get matrix W (diagonal matrix)
      w.element <- pi*(1-pi)
      colnames(w.element) <- "w.element"
      
      ## Step 2: Derivatives ##
      
      ## First
      # second part of first derivative
      data.w.pi <- cbind(data, pi, w.element)
      
      cell.sum.pi.withname <- ddply(data.w.pi, .variables=c("cell.name"),
                                    function(subdata, colnames){colSums(subdata[colnames])},
                                    "pi")
      colnames(cell.sum.pi.withname)[2:ncol(cell.sum.pi.withname)] <- 
        paste("cell.sum", "pi", sep=".")
      cell.sum.pi <- cell.sum.pi.withname[2:ncol(cell.sum.pi.withname)]
      
      d1e2_top <- as.matrix(cell.sum.pi)
      d1e2_bot <- t(Z) %*% pi
      
      # complete first derivative
      d1_top <- d1e1_top - d1e2_top
      d1_bot <- d1e1_bot - d1e2_bot
      
      ## Second
      
      # top-left element of second derivative (diagonal matrix)
      cell.sum.Welm <- ddply(data.w.pi, .variables = c("cell.name"),
                             function(dfr, colnm){sum(dfr[, colnm])}, "w.element")
      colnames(cell.sum.Welm)[2] <- paste("cell.sum", "w.element", sep=".")
      
      TWT.diag <- cell.sum.Welm[,2]
      
      # top-right element of second derivative
      WelmZ <- t(t(Z) * as.vector(w.element))
      colnames(WelmZ) <- paste("weighted", colnames(Z), sep=".")
      data.pi.Z <- cbind(data.w.pi, WelmZ)
      
      cell.sums <- as.data.frame(cell.sum.Welm)
      
      for (z in colnames(WelmZ)){
        cell.sum.Wcov <- ddply(data.pi.Z, .variables = c("cell.name"),
                               function(dfr, colnm){sum(dfr[, colnm])}, z)
        colnames(cell.sum.Wcov)[2] <- paste("cell.sum", z, sep=".")
        cell.sums <- join(cell.sums, cell.sum.Wcov, by="cell.name")
      }
      
      tr2 <- as.matrix(cell.sums[, 3:ncol(cell.sums)])
      
      # bottom-left element of second derivative
      bl2 <- t(tr2)
      
      # bottom-right element of second derivative
      br2 <- t(WelmZ) %*% Z
      
      ## Step 3: Inversion of Second derivative ##
      TWTinv.diag <- 1/TWT.diag
      inv.tl2 <- diag(TWTinv.diag)
      
      # Schur complement
      schur <- br2 - bl2 %*% inv.tl2 %*% tr2 
      inv.schur <- solve(schur, tol=inv.tol)
      
      # top-left element of second derivative inverse
      tl2inv <- inv.tl2 + inv.tl2 %*% tr2 %*% inv.schur %*% bl2 %*% inv.tl2
      
      # top-right element of second derivative inverse
      tr2inv <- - inv.tl2 %*% tr2 %*% inv.schur
      
      # bottom-left element of second derivative inverse
      bl2inv <- - inv.schur %*% bl2 %*% inv.tl2
      
      # bottom-right element of second derivative inverse
      br2inv <- inv.schur
      
      ## Step 4: Update, including linesearch ##
      
      ## Starting values
      # set linesearch parameter
      linesearch <- 1
      # set linesearch iteration counter
      j <- 1
      
      ## Loop over linesearch
      while (j <= maxit.linesearch){
        # Kappa (TS parameter)
        kappa_update <- tl2inv %*% d1_top + tr2inv %*% d1_bot
        kappa_new <- kappa + linesearch*kappa_update
        colnames(kappa_new) <- "kappa"
        
        # zeta (covariate parameter)
        zeta_update <- bl2inv %*% d1_top + br2inv %*% d1_bot
        zeta_new <- zeta + linesearch*zeta_update
        
        # mu (individual predictor)
        kappa_new.withname <- cbind(cell.sum.Y.withname[1], kappa_new)
        indiv.kappa_new <- join(indiv.cell.name, kappa_new.withname, by="cell.name") 
        
        mu_new <- indiv.kappa_new$kappa + Z %*% zeta_new
        
        # evaluate the likelihood at new values
        lik_new <- (t(Y) %*% mu_new) - sum(log(1+exp(mu_new)))
        
        # compare likelihoods: only to see if need to re-enter linesearch
        if (lik_new - lik < 0){
          # New likelihood is lower, have overstepped. Enter linesearch
          linesearch <- linesearch/2
          j <- j+1
        } else break
      }
      
      ## Update pi and first derivative
      
      # get Nx1 vector of probabilities
      pi_new <- exp(mu_new)/(1+exp(mu_new))
      colnames(pi_new) <- "pi"
      
      # get matrix W (diagonal matrix)
      w.element_new <- pi_new*(1-pi_new)
      colnames(w.element_new) <- "w.element"
      
      # get new second element of first derivative
      data.w.pi_new <- cbind(data, pi_new)
      
      cell.sum.pi_new.withname <- ddply(data.w.pi_new, .variables=c("cell.name"),
                                        function(subdata, colnames){colSums(subdata[colnames])},
                                        "pi")
      colnames(cell.sum.pi_new.withname)[2:ncol(cell.sum.pi_new.withname)] <- 
        paste("cell.sum", "pi_new", sep=".")
      cell.sum.pi_new <- cell.sum.pi_new.withname[2:ncol(cell.sum.pi_new.withname)]
      
      d1e2_top_new <- as.matrix(cell.sum.pi_new)
      d1e2_bot_new <- t(Z) %*% pi_new
      
      # complete first derivative
      d1_top_new <- d1e1_top - d1e2_top_new
      d1_bot_new <- d1e1_bot - d1e2_bot_new
      d1_new <- c(d1_top_new, d1_bot_new)    
      norm.d1 <- norm(as.matrix(d1_new), type="F")
      
      ## Step 5: Compare results of updating ##
      # Case 1: difference exceeds tolerance, update and re-enter loop
      if (lik_new - lik > tolerance){
        kappa <- kappa_new
        zeta <- zeta_new
        mu <- mu_new
        lik <- lik_new
        i <- i+1
        rm(kappa_new, zeta_new, mu_new, lik_new, pi_new, w.element_new,
           d1e2_top_new, d1e2_bot_new, d1_top_new, d1_bot_new)
        result <- "exceed lik tolerance, re-enter loop"
        r2 <- "at likelihood"
      } else if (lik_new - lik < 0){
        # Case 2: old likelihood higher, linesearch unsuccessful, terminate
        print(paste("overstepped and linesearch iteration limit reached,
                    exiting after", i-1, "iterations", sep=" "))
        i <- maxit.loop+4   # sets i too high for loop to continue
        result <- "overstep"
      } else if (lik_new - lik <= tolerance & lik_new - lik >= 0){
        # Case 3: likelihood condition satisfied, check first derivative
        # Case 3a: first derivative condition not satisfied, re-enter loop
        if (norm.d1 > d1.tol){
          kappa <- kappa_new
          zeta <- zeta_new
          mu <- mu_new
          lik <- lik_new
          i <- i+1
          rm(kappa_new, zeta_new, mu_new, lik_new, pi_new, w.element_new,
             d1e2_top_new, d1e2_bot_new, d1_top_new, d1_bot_new)
          result <- "exceed d1 tolerance, re-enter loop"
          r2 <- "at first derivative"
        } 
        # Case 3b: first derivative condition satisfied, terminate
        if (norm.d1 <= d1.tol){
          print(paste("converged after", i, "iterations", sep=" "))
          i <- maxit.loop+4 # sets i too high for loop to continue
          result <- "converge"
        }
      }
      # end of loop
    }
    
    if (i==maxit.loop+1)
      print(paste("max iterations exceeded, did not converge", r2, sep=" "))
    # at this point one ought to use the final values as the estimates and
    # also construct the SE based on these, but in the interests of expedience
    # I don't
    
    variance <- c(TWT.diag, diag(br2))
    Std.error <- sqrt(variance)
    
    coefficients.TS <- matrix(nrow=nrow(kappa), ncol=4)
    coefficients.TS[,1] <- kappa
    coefficients.TS[,2] <- Std.error[1:nrow(kappa)]
    coefficients.TS[,3] <- coefficients.TS[,1]/coefficients.TS[,2]
    coefficients.TS[,4] <- 2*pnorm(abs(coefficients.TS[,3]), lower.tail=FALSE)
    colnames(coefficients.TS) <- c("Estimate", "Std. Error", "z value", "Pr(>|z|)")
    rownames(coefficients.TS) <- cell.sum.Welm[,1]
    
    coefficients.covariates <- matrix(nrow=nrow(zeta), ncol=4)
    coefficients.covariates[,1] <- zeta
    coefficients.covariates[,2] <- Std.error[(nrow(kappa)+1):nrow(as.matrix(Std.error))]
    coefficients.covariates[,3] <- coefficients.covariates[,1]/coefficients.covariates[,2]
    coefficients.covariates[,4] <- 2*pnorm(abs(coefficients.covariates[,3]), lower.tail=FALSE)
    colnames(coefficients.covariates) <- c("Estimate", "Std. Error", "z value", "Pr(>|z|)")
    rownames(coefficients.covariates) <- rownames(zeta)
    }
  #### Case without covariates ####
  else{
    
    #### Calculate things that will be used repeatedly ####
    
    ## get Y and Z
    Y <- as.matrix(data[dep.var])
    
    ## get first part of first derivative
    # add cell.name to data: age-cohort cell identifier (if not already there)
    if (!"cell.name" %in% colnames(data)) {
      cell.name <- paste("ik", as.character(data$age), as.character(data$cohort),
                         sep="_")
      data <- cbind(cell.name, data)
    } else {
      warning("Variable cell.name must be unique identifier of age-cohort cells;
              if preexisting variable cell.name has alternate meaning please 
              rename")
    }
    
    cell.sum.Y.withname <- ddply(data, .variables=c("cell.name"),
                                 function(subdata, colnames){colSums(subdata[colnames])},
                                 dep.var)
    colnames(cell.sum.Y.withname)[2:ncol(cell.sum.Y.withname)] <- 
      paste("cell.sum", dep.var, sep=".")
    cell.sum.Y <- cell.sum.Y.withname[2:ncol(cell.sum.Y.withname)]
    
    d1e1 <- as.matrix(cell.sum.Y)
    
    #### Get initial starting values ####
    
    if((!is.null(custom.kappa)|!is.null(custom.zeta)) & !(init=="custom"))
      warning("custom.kappa and custom.zeta are not used without init=custom")
    
    if (init == "ols"){
      # use estimate.TS once to get estimated kappa and zeta
      initial.TS <- apc.indiv.estimate.TS(data, dep.var = dep.var)
      kappa <- as.matrix(initial.TS$kappahat)
    } else if (init == "zero"){
      kappa <- rep(0, nrow(cell.sum.Y))
    } else if (init == "custom"){
      kappa <- custom.kappa
    } else stop("init must be one of 'ols', 'zero', 'custom'")
    
    # set loop iteration counter
    i <- 1
    
    # get mu (individual predictor)
    kappa.withname <- cbind(cell.sum.Y.withname[1], kappa)
    indiv.cell.name <- data["cell.name"]
    indiv.kappa <- join(indiv.cell.name, kappa.withname, by="cell.name") 
    
    mu <- as.matrix(indiv.kappa$kappa) 
    colnames(mu) <- "mu"
    
    # evaluate the log-likelihood at the estimated values of kappa and zeta: 
    lik <- (t(Y) %*% mu) - sum(log(1+exp(mu)))
    
    # --------------------------
    
    #### The loop ####
    
    while (i <= maxit.loop){
      
      ## Step 1: Elements of updating ##
      
      # get Nx1 vector of probabilities
      pi <- exp(mu)/(1+exp(mu))
      colnames(pi) <- "pi"
      
      # get matrix W (diagonal matrix)
      w.element <- pi*(1-pi)
      colnames(w.element) <- "w.element"
      
      ## Step 2: Derivatives ##
      
      ## First
      data.w.pi <- cbind(data, pi, w.element)
      
      cell.sum.pi.withname <- ddply(data.w.pi, .variables=c("cell.name"),
                                    function(subdata, colnames){colSums(subdata[colnames])},
                                    "pi")
      colnames(cell.sum.pi.withname)[2:ncol(cell.sum.pi.withname)] <- 
        paste("cell.sum", "pi", sep=".")
      cell.sum.pi <- cell.sum.pi.withname[2:ncol(cell.sum.pi.withname)]
      
      d1e2 <- as.matrix(cell.sum.pi)
      
      # complete first derivative
      d1 <- d1e1 - d1e2
      
      ## Second
      # this is the "top-left" element in the covariate case
      cell.sum.Welm <- ddply(data.w.pi, .variables = c("cell.name"),
                             function(dfr, colnm){sum(dfr[, colnm])}, "w.element")
      colnames(cell.sum.Welm)[2] <- paste("cell.sum", "w.element", sep=".")
      
      d2.diag <- cell.sum.Welm[, 2]
      
      ## Step 3: Inversion of Second derivative ##
      
      # Invert top-left of second derivative
      d2inv.diag <- 1/d2.diag
      inv.d2 <- diag(d2inv.diag)
      
      ## Step 4: Update, including linesearch ##
      
      ## Starting values
      # set linesearch parameter
      linesearch <- 1
      # set linesearch iteration counter
      j <- 1
      
      ## Loop over linesearch
      while (j <= maxit.linesearch){
        # Kappa (TS parameter)
        kappa_update <- inv.d2 %*% d1 
        kappa_new <- kappa + linesearch*kappa_update
        colnames(kappa_new) <- "kappa"
        
        # get mu (individual predictor)
        kappa_new.withname <- cbind(cell.sum.Y.withname[1], kappa_new)
        indiv.kappa_new <- join(indiv.cell.name, kappa_new.withname, by="cell.name") 
        
        mu_new <- as.matrix(indiv.kappa_new$kappa) 
        
        # evaluate the likelihood at new values
        lik_new <- (t(Y) %*% mu_new) - sum(log(1+exp(mu_new)))
        
        # compare likelihoods: only to see if need to re-enter linesearch
        if (lik_new - lik < 0){
          # New likelihood is lower, have overstepped. Enter linesearch
          linesearch <- linesearch/2
          j <- j+1
        } else break
      }
      
      ## Update pi and first derivative
      
      # get Nx1 vector of probabilities
      pi_new <- exp(mu_new)/(1+exp(mu_new))
      colnames(pi_new) <- "pi"
      
      # get matrix W (diagonal matrix)
      w.element_new <- pi_new*(1-pi_new)
      colnames(w.element_new) <- "w.element"
      
      # get first derivative
      data.w.pi_new <- cbind(data, pi_new)
      
      cell.sum.pi_new.withname <- ddply(data.w.pi_new, .variables=c("cell.name"),
                                        function(subdata, colnames){colSums(subdata[colnames])},
                                        "pi")
      colnames(cell.sum.pi_new.withname)[2:ncol(cell.sum.pi_new.withname)] <- 
        paste("cell.sum", "pi_new", sep=".")
      cell.sum.pi_new <- cell.sum.pi_new.withname[2:ncol(cell.sum.pi_new.withname)]
      
      d1e2_new <- as.matrix(cell.sum.pi_new)
      
      # complete first derivative
      d1_new <- d1e1 - d1e2_new
      norm.d1 <- norm(as.matrix(d1_new), type="F")
      
      ## Step 5: Compare results of updating ##
      # Case 1: difference exceeds tolerance, update and re-enter loop
      if (lik_new - lik > tolerance){
        kappa <- kappa_new
        mu <- mu_new
        lik <- lik_new
        i <- i+1
        rm(kappa_new, mu_new, lik_new, pi_new, w.element_new, 
           d1e2_new)
        result <- "exceed lik tolerance, re-enter loop"
        r2 <- "at likelihood"
      } else if (lik_new - lik < 0){
        # Case 2: old likelihood higher, linesearch unsuccessful, terminate
        print(paste("overstepped and linesearch iteration limit reached,
                    exiting after", i-1, "iterations", sep=" "))
        i <- maxit.loop+4   # sets i too high for loop to continue
        result <- "overstep"
      } else if (lik_new - lik <= tolerance & lik_new - lik >= 0){
        # Case 3: likelihood condition satisfied, check first derivative
        # Case 3a: first derivative condition not satisfied, re-enter loop
        if (norm.d1 > d1.tol){
          kappa <- kappa_new
          mu <- mu_new
          lik <- lik_new
          i <- i+1
          rm(kappa_new, mu_new, lik_new, pi_new, w.element_new, 
             d1e2_new)
          result <- "exceed d1 tolerance, re-enter loop"
          r2 <- "at first derivative"
        } 
        # Case 3b: first derivative condition satisfied, terminate
        if (norm.d1 <= d1.tol){
          print(paste("converged after", i, "iterations", sep=" "))
          i <- maxit.loop+4 # sets i too high for loop to continue
          result <- "converge"
        }
      }
    }
    if (i==maxit.loop+1)
      print(paste("max iterations exceeded, did not converge", r2, sep=" "))
    # at this point one ought to use the final values as the estimates and
    # also construct the SE based on these, but in the interests of expedience
    # I don't
    
    variance <- d2inv.diag
    Std.error <- sqrt(variance)
    
    zeta <- NULL
    
    coefficients.TS <- matrix(nrow=nrow(kappa), ncol=4)
    coefficients.TS[,1] <- kappa
    coefficients.TS[,2] <- Std.error[1:nrow(kappa)]
    coefficients.TS[,3] <- coefficients.TS[,1]/coefficients.TS[,2]
    coefficients.TS[,4] <- 2*pnorm(abs(coefficients.TS[,3]), lower.tail=FALSE)
    colnames(coefficients.TS) <- c("Estimate", "Std. Error", "z value", "Pr(>|z|)")
    rownames(coefficients.TS) <- cell.sum.Welm[,1]
    
    coefficients.covariates <- NULL
    }
  
  # --------------------------
  #### Return objects of interest ####
  
  #Define
  param.dimension <- length(kappa)+length(zeta)
  aic <- -2*lik + 2*param.dimension
  
  #Deleted values
  if(!exists("lik_new")) lik_new <- NULL
  if(!exists("kappa_new")) kappa_new <- NULL
  if(!exists("zeta_new")) zeta_new <- NULL
  
  valuables <- list(likelihood = lik,
                    kappa = kappa,
                    zeta = zeta,
                    n.loop.iterations = i-1,
                    n.linesearch.iterations = j-1,
                    new.likelihood = lik_new,
                    new.kappa = kappa_new,
                    new.zeta = zeta_new,
                    param.dimension = param.dimension,
                    aic = aic,
                    std.error = Std.error,
                    coefficients.TS = coefficients.TS,
                    coefficients.covariates = coefficients.covariates,
                    d1_new = d1_new,
                    norm.d1 = norm.d1,
                    result = result)
  
  return (valuables)
  }